import unittest
import tempfile
import os
import shutil
import json
import csv
import warnings
from surquest.utils.appstoreconnect.analyticsreports.handler import Handler
from surquest.utils.appstoreconnect.analyticsreports.errors import (
    PayloadFormatError,
    NoValidIdsError,
    NoValidUrlsError,
)


class TestHandler(unittest.TestCase):

    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    def test_extract_ids_valid(self):
        payload = [{"id": "123"}, {"id": "456"}, {"id": "789"}]
        result = Handler.extract_ids(payload)
        assert result == ["123", "456", "789"]

    def test_extract_ids_with_invalid_items(self):
        payload = [{"id": "123"}, {"no_id": "abc"}, None, {"id": None}]
        with warnings.catch_warnings(record=True) as w:
            result = Handler.extract_ids(payload)
            assert result == ["123"]
            assert len(w) == 3  # Warnings for 3 invalid items

    def test_extract_ids_no_valid_ids(self):
        payload = [{"id": None}, {"foo": "bar"}]
        try:
            Handler.extract_ids(payload)
            assert False, "Expected NoValidIdsError to be raised"
        except NoValidIdsError as e:
            assert str(e) == "No valid IDs found in the payload."

    def test_extract_attribute_values_valid(self):
        payload = [
            {"attributes": {"processingDate": "2025-07-27"}},
            {"attributes": {"processingDate": "2025-07-28"}},
        ]
        result = Handler.extract_attribute_values(payload, attribute="processingDate")
        assert result == ["2025-07-27", "2025-07-28"]

    def test_extract_attribute_values_missing_attribute(self):
        payload = [{"attributes": {"other": "value"}}, {"attributes": {}}]
        with warnings.catch_warnings(record=True) as w:
            try:
                Handler.extract_attribute_values(payload, attribute="processingDate")
                assert False, "Expected NoValidUrlsError to be raised"
            except NoValidUrlsError as e:
                assert "No valid `processingDate`" in str(e)
            assert len(w) == 2

    def test_extract_attribute_values_without_attribute(self):
        payload = [
            {"attributes": {"key": "value"}},
            {"attributes": {"another_key": "another_value"}},
        ]
        result = Handler.extract_attribute_values(payload)
        assert result == [{"key": "value"}, {"another_key": "another_value"}]

    def test_deduplicate_data(self):
        data = [
            {"id": "123", "name": "Alice", "age": 30.5},
            {"id": "456", "name": "Bob", "age": "25"},
            {"id": "123", "name": "Alice", "age": "30.5"},
            {"id": "123", "name": "Alice", "age": "30.5"},
            {"id": "456", "name": "Bob", "age": 25},
        ]
        result = Handler.deduplicate_data(data)
        assert isinstance(result, list)
        assert len(result) == 2
        assert {"id": 123, "name": "Alice", "age": 30.5} in result
        assert {"id": 456, "name": "Bob", "age": 25} in result

    def test_create_directory_creates_dir(self):
        nested_dir = os.path.join(self.temp_dir, "subdir1", "subdir2")
        file_path = os.path.join(nested_dir, "file.txt")

        assert not os.path.exists(nested_dir)

        Handler.create_directory(file_path)

        assert os.path.exists(nested_dir)
        assert os.path.isdir(nested_dir)

    def test_create_directory_existing_dir(self):
        existing_dir = os.path.join(self.temp_dir, "existing_dir")
        os.makedirs(existing_dir)

        file_path = os.path.join(existing_dir, "file.txt")
        # Should not raise or fail if directory exists
        Handler.create_directory(file_path)
        assert os.path.exists(existing_dir)

    def test_list_of_dicts_to_jsonl_creates_file_with_correct_content(self):
        data = [
            {"a": 1, "b": "x"},
            {"a": 2, "b": "y"},
        ]
        file_path = os.path.join(self.temp_dir, "output.jsonl")

        Handler.list_of_dicts_to_jsonl(data, file_path)

        assert os.path.isfile(file_path)

        with open(file_path, "r", encoding="utf-8") as f:
            lines = f.read().splitlines()
            assert len(lines) == 2
            for i, line in enumerate(lines):
                obj = json.loads(line)
                assert obj == data[i]

    def test_list_of_dicts_to_csv_creates_file_with_correct_content(self):
        data = [
            {"name": "Alice", "age": 30},
            {"name": "Bob", "age": 25},
        ]
        file_path = os.path.join(self.temp_dir, "output.csv")

        Handler.list_of_dicts_to_csv(data, file_path)

        assert os.path.isfile(file_path)

        with open(file_path, newline="", encoding="utf-8") as f:
            reader = csv.DictReader(f)
            rows = list(reader)
            assert len(rows) == 2
            assert rows[0]["name"] == "Alice"
            assert rows[0]["age"] == "30"  # CSV stores all as strings
            assert rows[1]["name"] == "Bob"
            assert rows[1]["age"] == "25"

    def test_list_of_dicts_to_csv_empty_data_raises_value_error(self):
        data = []
        file_path = os.path.join(self.temp_dir, "empty.csv")

        try:
            Handler.list_of_dicts_to_csv(data, file_path)
        except ValueError as e:
            assert "data list is empty" in str(e)
        else:
            assert False, "ValueError not raised for empty data"

    def test_get_distinct_values_for_attribute(self):
        
        data = [
            {'id': 1, 'name': 'Alice', 'country': 'USA'},
            {'id': 2, 'name': 'Bob', 'country': 'UK'},
            {'id': 3, 'name': 'Alice', 'country': 'USA'},
            {'id': 4, 'name': 'David', 'country': None},
        ]

        result = Handler.get_distinct_values(data, 'country')
        assert sorted(result) == sorted(['USA', 'UK'])

    def test_filter_list_of_dicts(self):

        data = [
            {'id': 1, 'name': 'Alice', 'country': 'USA'},
            {'id': 2, 'name': 'Bob', 'country': 'UK'},
            {'id': 3, 'name': 'Alice', 'country': 'USA'},
            {'id': 4, 'name': 'David', 'country': None},
        ]

        expected = [
            {'id': 2, 'name': 'Bob', 'country': 'UK'}
        ]

        result = Handler.filter_list_of_dicts(
            data=data, 
            attribute='country', 
            value='UK'
        )
        
        assert result == expected


#!/usr/bin/env python3
# Unit tests for the Credentials class

import unittest
from surquest.utils.appstoreconnect.credentials import Credentials
import jwt
from unittest.mock import patch
from pathlib import Path
import datetime


class TestCredentials(unittest.TestCase):
    def setUp(self):
        self.issuer_id = "TEST_ISSUER_ID"
        self.key_id = "TEST_KEY_ID"
        # Sample ES256 private key (test-only)
        self.key_path = Path.cwd() / "credentials" / "key.p8"
        self.private_key = self.key_path.read_text()
        self.credentials = Credentials(
            issuer_id=self.issuer_id,
            key_id=self.key_id,
            private_key=self.private_key
        )

    def test_generate_token_valid(self):
        token = self.credentials.generate_token()
        decoded = jwt.decode(
            token,
            options={"verify_signature": False},  # We only check payload structure here
            algorithms=["ES256"]
        )

        assert decoded["iss"] == self.issuer_id
        assert decoded["aud"] == "appstoreconnect-v1"
        assert "iat" in decoded
        assert "exp" in decoded
        assert decoded["exp"] > decoded["iat"]

    def test_generate_token_expiration_limit(self):
        try:
            self.credentials.generate_token(expiration_minutes=60)
            assert False, "Expected ValueError for expiration > 20 minutes"
        except ValueError as e:
            assert "oken expiration must be between 1 and 20 minutes." in str(e)

    def test_default_expiration_is_20_minutes(self):
        token = self.credentials.generate_token()
        decoded = jwt.decode(
            token,
            options={"verify_signature": False},
            algorithms=["ES256"]
        )
        iat = decoded["iat"]
        exp = decoded["exp"]
        assert exp - iat == 20 * 60, \
            f"Difference between expected and actual expiration is {exp - iat} seconds"


import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Prevent messages from propagating to the root logger
logger.propagate = False

# Add handler only if this logger has no direct handlers
if not logger.handlers:
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(levelname)-8s - %(asctime)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)


class PayloadFormatError(ValueError):
    """Raised when the payload format is invalid (e.g. missing or malformed 'data' key)."""
    pass


class NoValidIdsError(ValueError):
    """Raised when no valid IDs are found in the payload."""
    pass


class NoValidUrlsError(ValueError):
    """Raised when no valid URLs are found in the payload."""
    pass



import unittest
import logging
from surquest.utils.appstoreconnect.analyticsreports.logger import logger


class TestLoggerSetup(unittest.TestCase):

    def test_logger_level_is_debug(self):
        assert logger.level == logging.DEBUG

    def test_logger_has_one_handler(self):
        assert len(logger.handlers) == 1
        assert isinstance(logger.handlers[0], logging.StreamHandler)

    def test_logger_formatter_format(self):
        handler = logger.handlers[0]
        fmt = handler.formatter._fmt
        expected_fmt = '%(levelname)-8s - %(asctime)s - %(message)s'
        assert fmt == expected_fmt

    def test_logger_propagation_disabled(self):
        assert logger.propagate is False

    def test_duplicate_handler_prevention(self):
        initial_count = len(logger.handlers)

        # Simulate running the initialization code again
        if not logger.handlers:
            handler = logging.StreamHandler()
            formatter = logging.Formatter('%(levelname)-8s - %(asctime)s - %(message)s')
            handler.setFormatter(formatter)
            logger.addHandler(handler)

        # Should not increase handler count
        assert len(logger.handlers) == initial_count


import jwt
from datetime import datetime, timedelta
from typing import Optional

class Credentials:
    """Handles creation and signing of JWT tokens for the App Store Connect API."""

    MAX_EXPIRATION_MINUTES = 20

    def __init__(self, issuer_id: str, key_id: str, private_key: str):
        """
        Initialize the Credentials object.

        Args:
            issuer_id (str): The issuer ID from App Store Connect.
            key_id (str): The key ID for the API key.
            private_key (str): The private key (P8 format) used for signing the token.
        """
        if not issuer_id or not key_id or not private_key:
            raise ValueError("issuer_id, key_id, and private_key are all required.")
        
        self.issuer_id = issuer_id
        self.key_id = key_id
        self.private_key = private_key

    def generate_token(self, expiration_minutes: int = 20) -> str:
        """
        Generate a signed JWT token.

        Args:
            expiration_minutes (int, optional): Token lifetime in minutes (max 20). Defaults to 20.

        Returns:
            str: The signed JWT token string.

        Raises:
            ValueError: If expiration exceeds the allowed maximum.
        """
        if not (0 < expiration_minutes <= self.MAX_EXPIRATION_MINUTES):
            raise ValueError(f"Token expiration must be between 1 and {self.MAX_EXPIRATION_MINUTES} minutes.")

        now = datetime.utcnow()
        exp_time = now + timedelta(minutes=expiration_minutes)

        payload = {
            "iss": self.issuer_id,
            "iat": int(now.timestamp()),
            "exp": int(exp_time.timestamp()),
            "aud": "appstoreconnect-v1"
        }

        headers = {
            "alg": "ES256",
            "kid": self.key_id,
            "typ": "JWT"
        }

        return jwt.encode(
            payload=payload,
            key=self.private_key,
            algorithm="ES256",
            headers=headers
        )



import requests
import warnings
from typing import Dict, Any, Optional, List, Set
import csv
import gzip
import io

from ..credentials import Credentials
from .handler import Handler
from .enums.category import Category
from .enums.granularity import Granularity
from .enums.report_name import ReportName
from .logger import logger
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


class APIClientError(Exception):
    """Custom exception for API client errors."""

    pass


class Client:
    """
    A client for interacting with the Apple AppStore Connect Analytics Report API.
    """

    BASE_URL = "https://api.appstoreconnect.apple.com/v1"

    def __init__(self, credentials: Credentials):
        """
        Initializes the API client.

        Args:
            credentials (Credentials): An instance of a credentials class
                                       that provides a `generate_token` method.
        """
        self.credentials = credentials
        self.session = requests.Session()
        self._configure_retries()
        logger.info("Initialized Client with provided credentials")

    def _configure_retries(self):
        """Configures retries for HTTP requests to handle transient errors."""
        retry_strategy = Retry(
            total=3,
            backoff_factor=1,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["GET"],
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        self.session.mount("https://", adapter)

    def _get_headers(self) -> Dict[str, str]:
        """Generates the authorization headers for API requests."""
        token = self.credentials.generate_token()
        return {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
        }

    def _get_request(
        self, url: str, params: Optional[Dict[str, Any]] = None
    ) -> Optional[Dict[str, Any]]:
        """Performs a GET request and returns JSON response or None."""
        headers = self._get_headers()
        logger.debug(f"GET {url} | Params: {params}")
        try:
            response = self.session.get(url, headers=headers, params=params)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.HTTPError as e:
            logger.error(f"HTTP Error: {e.response.status_code} - {e.response.text}")
        except requests.exceptions.RequestException as e:
            logger.error(f"Request Error: {e}")
        return None

    def _get_resource(
        self, resource_path: str, params: Optional[Dict[str, Any]] = None
    ) -> Optional[Dict[str, Any]]:
        """Builds URL and fetches resource."""
        return self._get_request(f"{self.BASE_URL}/{resource_path}", params)

    def _paginate(
        self, resource_path: str, params: Optional[Dict[str, Any]] = None
    ) -> List[Dict[str, Any]]:
        """Handles pagination and returns full list of data items."""
        results = []
        url = f"{self.BASE_URL}/{resource_path}"
        while url:
            response = self._get_request(url, params)
            if response and "data" in response:
                results.extend(response["data"])
            url = response.get("links", {}).get("next")
            params = None  # subsequent pages include params in URL
        return results

    # ----------------- Public API Methods -----------------

    def read_report_requests(
        self, app_id: str, params: Optional[Dict[str, Any]] = None
    ) -> List[Dict[str, Any]]:
        return self._paginate(f"apps/{app_id}/analyticsReportRequests", params)

    def read_report_for_specific_request(
        self, request_id: str, params: Optional[Dict[str, Any]] = None
    ) -> List[Dict[str, Any]]:
        return self._paginate(f"analyticsReportRequests/{request_id}/reports", params)

    def read_list_of_instances_of_report(
        self, report_id: str, params: Optional[Dict[str, Any]] = None
    ) -> List[Dict[str, Any]]:
        return self._paginate(f"analyticsReports/{report_id}/instances", params)

    def read_segments_for_report(
        self, instance_id: str, params: Optional[Dict[str, Any]] = None
    ) -> List[Dict[str, Any]]:
        return self._paginate(
            f"analyticsReportInstances/{instance_id}/segments", params
        )

    def download_report_to_dicts(
        self, report_url: str, normalize: bool = True
    ) -> Optional[List[Dict[str, str]]]:
        """Downloads a gzipped CSV report and parses it into a list of dictionaries."""
        try:
            csv_content = self._download_gzipped_csv(report_url)
            data = self._parse_csv_to_dicts(csv_content, normalize)
            out = []
            for item in data:
                # item.update({"url": report_url})
                out.append(item)
            return out
        except Exception:
            logger.exception("Failed to download or parse report")
            return None

    # ----------------- Helper Methods -----------------

    def _download_gzipped_csv(self, url: str) -> str:
        """Downloads gzipped CSV and returns as string."""
        response = self.session.get(
            url, headers={"Accept-Encoding": "gzip"}, stream=True
        )
        response.raise_for_status()
        decompressed_content = gzip.decompress(response.content)
        return decompressed_content.decode("utf-8")

    def _parse_csv_to_dicts(
        self, csv_content: str, normalize: bool
    ) -> List[Dict[str, str]]:
        """Parses CSV content into list of dictionaries."""
        csv_file = io.StringIO(csv_content)
        reader = csv.DictReader(csv_file, delimiter="\t")
        if normalize:
            return [
                {key.lower().replace(" ", "_").replace("-", "_"): value for key, value in row.items()}
                for row in reader
            ]
        return list(reader)

    def list_report_dates(
        self,
        report_name: ReportName,
        app_id: Optional[str] = None,
        report_ids: Optional[List[str]] = None,
        granularity: Granularity = Granularity.DAILY,
    ) -> List[str]:
        if not app_id and not report_ids:
            raise APIClientError("Either 'app_id' or 'report_ids' must be provided.")

        if not report_ids:
            response = self.list_reports(
                app_id=app_id, report_name=report_name, category=report_name.category
            )
            report_ids = Handler.extract_ids(response)

        dates: Set[str] = set()
        for report_id in report_ids:
            instances = self.read_list_of_instances_of_report(
                report_id, params={"filter[granularity]": granularity.value}
            )
            if not instances:
                raise APIClientError(f"No instances found for report: {report_id}")
            available_dates = Handler.extract_attribute_values(
                instances, attribute="processingDate"
            )
            if available_dates:
                dates.update(available_dates)
        return sorted(dates)

    def list_reports(
        self,
        app_id: str,
        category: Optional[Category] = None,
        report_name: Optional[ReportName] = None,
    ) -> List[Dict[str, Any]]:
        query_params = {}
        if category:
            query_params["filter[category]"] = category.value
        if report_name:
            query_params["filter[name]"] = report_name.value

        report_requests = self.read_report_requests(app_id)
        report_ids = Handler.extract_ids(report_requests)
        if not report_ids:
            raise APIClientError(f"No report requests found for app: {app_id}")

        reports = []
        for request_id in report_ids:
            reports.extend(
                self.read_report_for_specific_request(request_id, params=query_params)
            )
        return reports

    def get_data(
        self,
        app_id: str,
        report_name: ReportName,
        granularity: Granularity = Granularity.DAILY,
        dates: Optional[Set[str]] = None,
    ) -> List[Dict[str, str]]:
        dates = dates or set()
        data: List[Dict[str, str]] = []

        report_ids = self._fetch_report_ids(app_id, report_name)
        if not dates:
            dates = set(
                self.list_report_dates(
                    report_name, report_ids=report_ids, granularity=granularity
                )
            )

        instance_ids = self._fetch_instance_ids(report_ids, granularity, dates)
        urls = self._fetch_segment_urls(instance_ids)

        date_slices = dict()
        
        for url in urls: # URLs are sorted older are processed before newer
            segment_data = self.download_report_to_dicts(url)
        
            if segment_data:
                # -------------------------------------------------------------- #
                # Important: This part requires patch
                # - for each dataset we have to analyze set of avaialble dates
                #   and for each date we ahve to add them to report dates 
                #   or overwrite it in report_dates
                # -------------------------------------------------------------- #
                
                available_dates = Handler.get_distinct_values(
                    data=segment_data,
                    key='date'
                )

                for avaialble_date in available_dates:

                    data_slice = Handler.filter_list_of_dicts(
                        data=segment_data,
                        attribute='date',
                        value=avaialble_date,
                        comparator='=='
                    )

                    date_slices[avaialble_date] = data_slice

        for date, data_slice in date_slices.items():
            data.extend(data_slice)
        
        return Handler.deduplicate_data(data)

    def fetch_customer_reviews(
        self,
        app_id: str,
        last_known_customer_review_id: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
        max_iterations: Optional[int] = None,

    ) -> List[Dict[str, Any]]:
        """
        Fetches all customer reviews for a given app_id.

        Args:
            app_id (str): The ID of the app.
            last_known_customer_review_id (Optional[str]): If provided, stops pagination once this review ID is found.
            params (Optional[Dict[str, Any]]): Query parameters for the request (default: sort=-createdDate, limit=200).
            max_iterations (Optional[int]): Max number of pagination requests (for integration testing). Unlimited by default.

        Returns:
            List[Dict[str, Any]]: A list of customer review records.
        """
        results = []
        seen_ids = set()
        found_last_known = False
        iterations = 0

        query_params = {
            "limit": 200,
            "sort": "-createdDate",
        }
        if params:
            query_params.update(params)

        url = f"{self.BASE_URL}/apps/{app_id}/customerReviews"

        while url and not found_last_known:
            if max_iterations is not None and iterations >= max_iterations:
                logger.info(f"Reached max iteration limit: {max_iterations}")
                break

            response = self._get_request(url, query_params)
            if not response:
                break

            data = response.get("data", [])
            for item in data:
                review_id = item.get("id")
                if last_known_customer_review_id and review_id == last_known_customer_review_id:
                    found_last_known = True
                    break
                if review_id not in seen_ids:
                    results.append(item)
                    seen_ids.add(review_id)

            url = response.get("links", {}).get("next")
            query_params = None  # Only pass params on the first request
            iterations += 1

        return results

    # ----------------- Private Steps for get_data -----------------

    def _fetch_report_ids(self, app_id: str, report_name: ReportName) -> List[str]:
        reports = self.list_reports(
            app_id, category=report_name.category, report_name=report_name
        )
        return Handler.extract_ids(reports)

    def _fetch_instance_ids(
        self, report_ids: List[str], granularity: Granularity, dates: Set[str]
    ) -> List[str]:
        instance_ids: List[str] = []
        for report_id in report_ids:
            for date in dates:
                instances = self.read_list_of_instances_of_report(
                    report_id,
                    params={
                        "filter[granularity]": granularity.value,
                        "filter[processingDate]": date,
                    },
                )
                ids = Handler.extract_ids(instances)
                instance_ids.extend(ids)
        return instance_ids

    def _fetch_segment_urls(self, instance_ids: List[str]) -> List[str]:
        urls: List[str] = []
        for instance_id in instance_ids:
            segments = self.read_segments_for_report(instance_id)
            urls.extend(Handler.extract_attribute_values(segments, attribute="url"))
        return sorted(urls)



from enum import Enum
from .category import Category


class ReportName(str, Enum):
    AIRPLAY_DISCOVERY_SESSIONS = ("AirPlay Discovery Sessions",  Category.FRAMEWORK_USAGE)
    HOME_SCREEN_WIDGET_ROTATIONS = ("Home Screen Widget Rotations",  Category.FRAMEWORK_USAGE)
    FILE_BASED_VIDEO_PLAYBACK_USAGE = ("File-Based Video Playback Usage",  Category.FRAMEWORK_USAGE)
    SAFARI_EXTENSIONS_USAGE = ("Safari Extensions Usage",  Category.FRAMEWORK_USAGE)
    CRABS_BASED_VIDEO_PLAYBACK_USAGE = ("CRABS-Based Video Playback Usage",  Category.FRAMEWORK_USAGE)
    TRANSLATION_REQUEST_USAGE = ("Translation Request Usage",  Category.FRAMEWORK_USAGE)
    SAFARI_EXTENSIONS_ENABLEMENT = ("Safari Extensions Enablement",  Category.FRAMEWORK_USAGE)
    AUTOMATIC_SPEECH_RECOGNITION_USAGE = ("Automatic Speech Recognition Usage",  Category.FRAMEWORK_USAGE)
    KEYBOARD_DICTATION_USAGE = ("Keyboard Dictation Usage",  Category.FRAMEWORK_USAGE)
    SPATIAL_AUDIO_USAGE = ("Spatial Audio Usage",  Category.FRAMEWORK_USAGE)
    VIDEO_DURATION_INFORMATION = ("Video Duration Information",  Category.FRAMEWORK_USAGE)
    DEFAULT_BROWSER_USAGE_RATE = ("Default Browser Usage Rate",  Category.FRAMEWORK_USAGE)
    HOME_SCREEN_WIDGETS = ("Home Screen Widgets",  Category.FRAMEWORK_USAGE)
    BROWSER_CHOICE_SCREEN_ENGAGEMENT_IOS_VERSIONS_BEFORE_182 = ("Browser Choice Screen Engagement (iOS versions before 18.2)",  Category.FRAMEWORK_USAGE)
    HOME_SCREEN_WIDGET_USAGE = ("Home Screen Widget Usage",  Category.FRAMEWORK_USAGE)
    VISIONKIT_IMAGE_ANALYSIS = ("VisionKit Image Analysis",  Category.FRAMEWORK_USAGE)
    REMINDERS_USAGE = ("Reminders Usage",  Category.FRAMEWORK_USAGE)
    PHOTOKIT_IMPORTS = ("PhotoKit Imports",  Category.FRAMEWORK_USAGE)
    NETWORKING_CONNECTION_ACTIVITY = ("Networking Connection Activity",  Category.PERFORMANCE)
    SPEECH_FRAMEWORK_TRANSCRIPTION_REQUESTS = ("Speech Framework Transcription Requests",  Category.FRAMEWORK_USAGE)
    HTTP_LIVE_STREAMING_VIDEO_PLAYBACK_USAGE = ("HTTP Live Streaming Video Playback Usage",  Category.FRAMEWORK_USAGE)
    MODE_ACTIVITY_NOTIFICATIONS = ("Mode Activity Notifications",  Category.FRAMEWORK_USAGE)
    METAL_COMMAND_QUEUES = ("Metal Command Queues",  Category.FRAMEWORK_USAGE)
    LOCK_SCREEN_WIDGET_CONFIGURATION = ("Lock Screen Widget Configuration",  Category.FRAMEWORK_USAGE)
    LOCAL_NETWORK_PRIVACY = ("Local Network Privacy",  Category.FRAMEWORK_USAGE)
    PHOTOS_LIBRARY_ACCESS = ("Photos Library Access",  Category.FRAMEWORK_USAGE)
    GAME_CONTROLLER_SESSIONS = ("Game Controller Sessions",  Category.FRAMEWORK_USAGE)
    SPEECH_FRAMEWORK_TRANSCRIPTION_REQUEST_AUDIO_DURATION = ("Speech Framework Transcription Request Audio Duration",  Category.FRAMEWORK_USAGE)
    TEXT_INPUT_ACTIONS = ("Text-Input Actions",  Category.FRAMEWORK_USAGE)
    VISIONKIT_SESSIONS = ("VisionKit Sessions",  Category.FRAMEWORK_USAGE)
    CAMETALLAYER_PERFORMANCE = ("CAMetalLayer Performance",  Category.PERFORMANCE)
    BLUETOOTH_SYSTEM_WAKES = ("Bluetooth System Wakes",  Category.PERFORMANCE)
    APP_RUNTIME_USAGE = ("App Runtime Usage",  Category.FRAMEWORK_USAGE)
    APP_DOWNLOADS_STANDARD = ("App Downloads Standard",  Category.COMMERCE)
    APP_DOWNLOADS_DETAILED = ("App Downloads Detailed",  Category.COMMERCE)
    APP_INSTALL_PERFORMANCE = ("App Install Performance",  Category.PERFORMANCE)
    APP_STORE_INSTALLATION_AND_DELETION_STANDARD = ("App Store Installation and Deletion Standard",  Category.APP_USAGE)
    APP_STORE_INSTALLATION_AND_DELETION_DETAILED = ("App Store Installation and Deletion Detailed",  Category.APP_USAGE)
    APP_SESSIONS_STANDARD = ("App Sessions Standard",  Category.APP_USAGE)
    APP_SESSIONS_DETAILED = ("App Sessions Detailed",  Category.APP_USAGE)
    APP_STORE_PRE_ORDERS_STANDARD = ("App Store Pre-Orders Standard",  Category.COMMERCE)
    APP_STORE_PRE_ORDERS_DETAILED = ("App Store Pre-Orders Detailed",  Category.COMMERCE)
    APP_STORE_PURCHASES_STANDARD = ("App Store Purchases Standard",  Category.COMMERCE)
    APP_STORE_PURCHASES_DETAILED = ("App Store Purchases Detailed",  Category.COMMERCE)
    APP_STORE_DISCOVERY_AND_ENGAGEMENT_STANDARD = ("App Store Discovery and Engagement Standard",  Category.APP_STORE_ENGAGEMENT)
    APP_STORE_DISCOVERY_AND_ENGAGEMENT_DETAILED = ("App Store Discovery and Engagement Detailed",  Category.APP_STORE_ENGAGEMENT)
    FLASHLIGHT_USAGE = ("Flashlight Usage",  Category.FRAMEWORK_USAGE)
    ARKIT_WORLD_TRACKING = ("ARKit World Tracking",  Category.FRAMEWORK_USAGE)
    ARKIT_WORLD_TRACKING_IMAGE_DETECTION = ("ARKit World Tracking Image Detection",  Category.FRAMEWORK_USAGE)
    PHOTOGRAMMETRY_OBJECTCAPTURESESSION_API_USAGE = ("Photogrammetry ObjectCaptureSession API Usage",  Category.FRAMEWORK_USAGE)
    PHOTOGRAMMETRYSESSION_API_USAGE = ("PhotogrammetrySession API Usage",  Category.FRAMEWORK_USAGE)
    HAPTICS_ENGINE_USAGE = ("Haptics Engine Usage",  Category.FRAMEWORK_USAGE)
    SHAZAMKIT_USAGE = ("ShazamKit Usage",  Category.FRAMEWORK_USAGE)
    APP_ADDED_TO_FOCUS = ("App Added to Focus",  Category.FRAMEWORK_USAGE)
    STREAMING_PLAYBACK_PERFORMANCE = ("Streaming Playback Performance",  Category.PERFORMANCE)
    STREAMING_DOWNLOADS_PERFORMANCE = ("Streaming Downloads Performance",  Category.PERFORMANCE)
    AUDIO_INPUT_MUTING = ("Audio Input Muting",  Category.FRAMEWORK_USAGE)
    AIRPLAY_PERFORMANCE = ("AirPlay Performance",  Category.PERFORMANCE)
    ROOMPLAN_USAGE = ("RoomPlan Usage",  Category.FRAMEWORK_USAGE)
    AUDIO_INPUT_ROUTE_AND_DURATION_AND_CALL_MODE = ("Audio Input Route and Duration and Call Mode",  Category.FRAMEWORK_USAGE)
    APP_DISK_SPACE_USAGE = ("App Disk Space Usage",  Category.FRAMEWORK_USAGE)
    LOCATION_SESSIONS = ("Location Sessions",  Category.FRAMEWORK_USAGE)
    CARPLAY_NAVIGATION = ("CarPlay Navigation",  Category.FRAMEWORK_USAGE)
    VIDEO_PIP_DURATION = ("Video PiP Duration",  Category.FRAMEWORK_USAGE)
    HTTP_LIVE_STREAMING_PLAYBACK_ERRORS = ("HTTP Live Streaming Playback Errors",  Category.PERFORMANCE)
    HTTP_LIVE_STREAMING_PLAYBACK_COUNT = ("HTTP Live Streaming Playback Count",  Category.FRAMEWORK_USAGE)
    BLUETOOTH_LE_SCANS = ("Bluetooth LE Scans",  Category.FRAMEWORK_USAGE)
    BLUETOOTH_LE_ADVERTISING = ("Bluetooth LE Advertising",  Category.FRAMEWORK_USAGE)
    ACCESSORYSETUPKIT_USAGE = ("AccessorySetupKit Usage",  Category.FRAMEWORK_USAGE)
    APP_CLIP_USAGE_STANDARD = ("App Clip Usage Standard",  Category.APP_USAGE)
    APP_CLIP_USAGE_DETAILED = ("App Clip Usage Detailed",  Category.APP_USAGE)
    APP_CRASHES = ("App Crashes",  Category.APP_USAGE)
    CUSTOM_LANGUAGE_MODEL_BUILDS_FAILED = ("Custom Language Model Builds Failed",  Category.PERFORMANCE)
    CUSTOM_LANGUAGE_MODEL_BUILDS_STARTED = ("Custom Language Model Builds Started",  Category.FRAMEWORK_USAGE)
    CUSTOMIZED_TRANSCRIPTION_REQUESTS = ("Customized Transcription Requests",  Category.FRAMEWORK_USAGE)
    DOCKKIT_APP_USAGE = ("DockKit App Usage",  Category.FRAMEWORK_USAGE)
    LIVE_ACTIVITY_USE = ("Live Activity Use",  Category.FRAMEWORK_USAGE)
    MULTIPLE_GAME_CONTROLLERS_USAGE = ("Multiple Game Controllers Usage",  Category.FRAMEWORK_USAGE)
    PHOTOS_PICKER = ("Photos Picker",  Category.FRAMEWORK_USAGE)
    SHARED_WITH_YOU_CONTENT_ENGAGEMENT = ("Shared With You Content Engagement",  Category.FRAMEWORK_USAGE)
    SPOTLIGHT_QUERY_PERFORMANCE = ("Spotlight Query Performance",  Category.PERFORMANCE)
    IBEACON_ADD_REGION_USAGE = ("iBeacon Add Region Usage",  Category.FRAMEWORK_USAGE)
    IBEACON_STOP_MONITORING_FOR_REGION_USAGE = ("iBeacon Stop Monitoring for Region Usage",  Category.FRAMEWORK_USAGE)
    VISIONKIT_LIVE_TEXT_USAGE = ("VisionKit Live Text Usage",  Category.FRAMEWORK_USAGE)
    WIFI_KNOWN_NETWORK_MODIFICATIONS = ("Wi-Fi Known Network Modifications",  Category.FRAMEWORK_USAGE)
    ACCESSORYSETUPKIT_ACCESSORY_PICKER_SESSIONS = ("AccessorySetupKit Accessory Picker Sessions",  Category.FRAMEWORK_USAGE)
    BLUETOOTH_LE_SESSION_DURATION = ("Bluetooth LE Session Duration",  Category.PERFORMANCE)
    IBEACON_RANGING_USAGE = ("iBeacon Ranging Usage",  Category.FRAMEWORK_USAGE)
    BROWSER_CHOICE_SCREEN_SELECTION = ("Browser Choice Screen Selection",  Category.FRAMEWORK_USAGE)
    APP_STORE_WEB_PREVIEW_ENGAGEMENT_STANDARD = ("App Store Web Preview Engagement Standard",  Category.APP_STORE_ENGAGEMENT)
    APP_STORE_WEB_PREVIEW_ENGAGEMENT_DETAILED = ("App Store Web Preview Engagement Detailed",  Category.APP_STORE_ENGAGEMENT)
    PLATFORM_APP_INSTALLS = ("Platform App Installs",  Category.APP_USAGE)
    AIRPLAY_ERRORS = ("AirPlay Errors",  Category.PERFORMANCE)
    APP_OPT_IN = ("App Opt In",  Category.APP_USAGE)
    CORE_LOCATION_GEOFENCING = ("Core Location Geofencing",  Category.FRAMEWORK_USAGE)
    SHAREPLAY_USAGE_BY_ACTIVITY_TYPE = ("SharePlay Usage by Activity Type",  Category.FRAMEWORK_USAGE)
    NOTIFICATION_SUMMARY_ENGAGEMENT = ("Notification Summary Engagement",  Category.FRAMEWORK_USAGE)
    VISIONKIT_DATA_DETECTORS = ("VisionKit Data Detectors",  Category.FRAMEWORK_USAGE)
    CALL_SERVICES_AND_CALL_PERFORMANCE = ("Call Services and Call Performance",  Category.FRAMEWORK_USAGE)
    LOAD_COREML_MODELS_METRICS = ("Load CoreML Models Metrics",  Category.FRAMEWORK_USAGE)
    VERIFY_WITH_WALLET_DOCUMENT_REQUESTS = ("Verify with Wallet Document Requests",  Category.FRAMEWORK_USAGE)
    VERIFY_WITH_WALLET_DOCUMENT_REQUEST_AVAILABILITY = ("Verify With Wallet Document Request Availability",  Category.FRAMEWORK_USAGE)
    APP_SESSIONS_CONTEXT = ("App Sessions Context",  Category.FRAMEWORK_USAGE)
    CARPLAY_APP_USAGE = ("CarPlay App Usage",  Category.APP_USAGE)
    AUDIO_VOLUME_LEVELS_AND_DURATION = ("Audio Volume Levels and Duration",  Category.FRAMEWORK_USAGE)
    BLUETOOTH_LE_SESSIONS = ("Bluetooth LE Sessions",  Category.FRAMEWORK_USAGE)
    ARKIT_VIDEO_FORMATS = ("ARKit Video Formats",  Category.FRAMEWORK_USAGE)
    ARKIT_CAPTURE_FRAME_RATE_THROTTLING = ("ARKit Capture Frame Rate Throttling",  Category.FRAMEWORK_USAGE)
    ARKIT_ARSESSION_DURATION = ("ARKit ARSession Duration",  Category.FRAMEWORK_USAGE)
    ARKIT_ARSESSION_FAILURES = ("ARKit ARSession Failures",  Category.FRAMEWORK_USAGE)
    ARKIT_FACE_TRACKING = ("ARKit Face Tracking",  Category.FRAMEWORK_USAGE)
    ANIMOJI_STICKERS_SENT = ("Animoji Stickers Sent",  Category.FRAMEWORK_USAGE)
    DISPLAY_POWER_INFORMATION = ("Display Power Information",  Category.PERFORMANCE)
    CORE_LOCATION_AUTHORIZATION_RESULTS = ("Core Location Authorization Results",  Category.FRAMEWORK_USAGE)
    LAUNCH_IMAGE_OVER_MEMORY_LIMIT = ("Launch Image Over Memory Limit",  Category.PERFORMANCE)
    GAME_CONTROLLER_HAPTICS_ENGINE_CREATION = ("Game Controller Haptics Engine Creation",  Category.FRAMEWORK_USAGE)
    PHOTOS_SHARING = ("Photos Sharing",  Category.FRAMEWORK_USAGE)
    AUDIO_SESSION_AUDIO_UNIT_USAGE = ("Audio Session Audio Unit Usage",  Category.FRAMEWORK_USAGE)
    COLLABORATION_MESSAGE_USAGE = ("Collaboration Message Usage",  Category.FRAMEWORK_USAGE)
    DYNAMIC_ISLAND_LAYOUT_CHANGES = ("Dynamic Island Layout Changes",  Category.FRAMEWORK_USAGE)
    APP_CRASHES_EXPANDED = ("App Crashes Expanded",  Category.PERFORMANCE)
    APP_STORAGE_READS_AND_WRITES = ("App Storage Reads and Writes",  Category.PERFORMANCE)
    APPLICATION_PREFERRED_LANGUAGE_SETTINGS = ("Application Preferred Language Settings",  Category.FRAMEWORK_USAGE)
    ARKIT_COLLABORATIVE_SESSION_FEATURES = ("ARKit Collaborative Session Features",  Category.FRAMEWORK_USAGE)
    AUDIO_OVERLOADS = ("Audio Overloads",  Category.PERFORMANCE)
    BLUETOOTH_LE_CONNECTION_RESULTS = ("Bluetooth LE Connection Results",  Category.FRAMEWORK_USAGE)
    BLUETOOTH_LE_CONNECTIONS_PER_APP = ("Bluetooth LE Connections Per App",  Category.FRAMEWORK_USAGE)
    BLUETOOTH_LE_DISCONNECTION_RESULTS = ("Bluetooth LE Disconnection Results",  Category.FRAMEWORK_USAGE)
    HOME_SCREEN_WIDGET_INSTALLS = ("Home Screen Widget Installs",  Category.FRAMEWORK_USAGE)
    
    # None standard option for customer reviews fetched via https://api.appstoreconnect.apple.com/v1/apps/{app_id}/customerReviews"
    APP_CUSTOMER_REVIEWS = ("App Customer Reviews", None)

    def __new__(cls, value: str, category: Category) -> "ReportName":
        obj = str.__new__(cls, value)
        obj._value_ = value
        obj._category = category
        return obj

    @property
    def category(self) -> Category:
        return self._category


# Usage Example
# print(ReportType.APP_DOWNLOADS_STANDARD.value)   # "App Downloads Standard"
# print(ReportType.APP_DOWNLOADS_STANDARD.category)  # Category.COMMERCE


import json
from src.surquest.utils.appstoreconnect.credentials import Credentials
from src.surquest.utils.appstoreconnect.analyticsreports.client import Client
from src.surquest.utils.appstoreconnect.analyticsreports.handler import Handler
from src.surquest.utils.appstoreconnect.analyticsreports.enums.category import Category
from src.surquest.utils.appstoreconnect.analyticsreports.enums.granularity import Granularity
from src.surquest.utils.appstoreconnect.analyticsreports.enums.report_name import ReportName
from pathlib import Path

key_path = Path.cwd() / "credentials" / "key.p8"
CATEGORY = Category.FRAMEWORK_USAGE
REPORT_NAME = ReportName.APP_STORE_INSTALLATION_AND_DELETION_DETAILED
GRANULARITY = Granularity.DAILY
DATE = "2025-07-27"
APP_ID = "950949627"

KEY_ID = '5WDUV3USAU' # 10-character key ID from App Store Connect
ISSUER_ID = '69a6de80-fd44-47e3-e053-5b8c7c11a4d1'  # 36-character issuer ID
PRIVATE_KEY = key_path.read_text()

credentials = Credentials(
    issuer_id=ISSUER_ID, # 36-character issuer ID
    key_id=KEY_ID, # 10-character key ID from App Store Connect
    private_key=PRIVATE_KEY
)
client = Client(credentials=credentials)
# token = credentials.generate_token()
#1A Read Report Requests
# report_requests = client.read_report_requests(app_id=APP_ID)
# data = client.get_data(
#     app_id = APP_ID,
#     report_name = REPORT_NAME,
#     category = CATEGORY,
#     granularity = GRANULARITY,
#     # dates = {'2025-07-25', '2025-07-26'}
# )

data = client.list_reports(
    app_id = APP_ID
)

# Extract Report types
print(json.dumps(Handler.extract_attribute_values(data), indent=4))

dates = client.list_report_dates(
    report_name = REPORT_NAME,
    app_id = APP_ID,
    granularity = GRANULARITY
)

print(f"/{dates}/")
print(json.dumps(list(dates), indent=4))



import os
import csv
import json
import warnings
import operator
from typing import Any, List, Dict

from .errors import PayloadFormatError, NoValidIdsError, NoValidUrlsError
from .logger import logger


class Handler:
    @staticmethod
    def extract_ids(payload: dict) -> list[str]:
        """
        Extracts a list of 'id' values from a payload dictionary.

        Raises:
            PayloadFormatError: if 'data' is missing or not a list.
            NoValidIdsError: if no valid IDs are found in the payload.
        """

        data = payload

        ids = []
        for item in data:
            if not isinstance(item, dict):
                warnings.warn(f"Skipped: item is not a dictionary: {item}")
                continue
            if "id" not in item or item["id"] is None:
                warnings.warn(f"Skipped: missing or null 'id' in item: {item}")
                continue
            ids.append(item["id"])

        if not ids:
            raise NoValidIdsError("No valid IDs found in the payload.")

        return ids


    @staticmethod
    def extract_attribute_values(payload: dict, attribute: str | None = None) -> list[str]:
        """
        Extracts a list of 'attribute' values from each item's attributes in the payload.

        Args:
            payload (dict): Dictionary representing the payload.
            attribute (str, optional): Name of the attribute to extract. Defaults to None.

        Raises:
            PayloadFormatError: if 'data' is missing or not a list.
            NoValidUrlsError: if no valid URLs are found in the payload.
        """

        data = payload
        out = []
        for item in data:
            if not isinstance(item, dict):
                warnings.warn(f"Skipped: item is not a dictionary: {item}")
                continue

            attributes = item.get("attributes")
            if not isinstance(attributes, dict):
                warnings.warn(f"Skipped: missing or malformed 'attributes' in item: {item}")
                continue
            
            if attribute is not None:
                val = attributes.get(attribute)
                if not val:
                    warnings.warn(f"Skipped: missing or empty `{attribute}` in attributes: {attributes}")
                    continue
                out.append(val)

            else:
                out.append(attributes)

        if not out:
            raise NoValidUrlsError(f"No valid `{attribute}` found in the payload.")

        return out

    @staticmethod
    def deduplicate_data(data: list[dict]) -> list[dict]:
        """
        Remove duplicated entries (dictionaries) and ensure consistent key order.

        Args:
            data (list): List of dictionaries

        Returns:
            list: Deduplicated list of dictionaries with consistent key order
        """
        if not data:
            return []

        # Convert string representations of numbers to actual numbers
        for item in data:
            for key, value in item.items():
                if isinstance(value, str):
                    if "." in value:
                        try:
                            # Try converting to float first (handles integers too)
                            num_value = float(value)
                            item[key] = num_value
                        except ValueError:
                            # Not a numeric string, keep as is
                            pass
                    elif value in [""]:
                        item[key] = None
                    else:
                        try:
                            # Try converting to integer
                            num_value = int(value)
                            item[key] = num_value
                        except ValueError:
                            # Not an integer, keep as is
                            pass

        # Determine consistent key order (from the first dictionary)
        key_order = list(data[0].keys())

        # Deduplicate using tuple representation
        seen = set()
        unique_data = []
        for d in data:
            # Ensure dictionary has all keys (fill missing with None)
            normalized = tuple((k, d.get(k)) for k in key_order)
            if normalized not in seen:
                seen.add(normalized)
                unique_data.append({k: d.get(k) for k in key_order})

        logger.info(
            f"Entries: duplicated {len(data) - len(unique_data)}, "
            f"original: {len(data)}, "
            f"deduplicated: {len(unique_data)}"
        )

        return unique_data

    @staticmethod
    def create_directory(file_path: str) -> None:
        """
        Creates a directory if it doesn't exist.

        Args:
            file_path (str): Path to the file or directory.
        """

        directory = os.path.dirname(file_path)
        if not os.path.exists(directory):
            os.makedirs(directory)

    @staticmethod
    def list_of_dicts_to_jsonl(data: list[dict], file_path: str) -> None:
        """
        Converts a list of dictionaries to JSON Lines format and writes to a file.

        Args:
            data (list[dict]): List of dictionaries to convert.
            file_path (str): Path to the output .jsonl file.
        """

        # Create the directory if it doesn't exist
        Handler.create_directory(file_path)

        with open(file_path, 'w', encoding='utf-8') as f:
            for item in data:
                json_line = json.dumps(item, ensure_ascii=False)
                f.write(json_line + '\n')

    @staticmethod
    def list_of_dicts_to_csv(data: list[dict], file_path: str) -> None:
        """
        Converts a list of dictionaries to CSV format and writes to a file.

        Args:
            data (list[dict]): List of dictionaries to convert.
            file_path (str): Path to the output .csv file.
        """
        if not data:
            raise ValueError("The data list is empty.")

        # Create the directory if it doesn't exist
        Handler.create_directory(file_path)
        
        # Extract headers from keys of the first dictionary
        headers = data[0].keys()
        
        with open(file_path, 'w', newline='', encoding='utf-8') as f:
            writer = csv.DictWriter(f, fieldnames=headers)
            writer.writeheader()
            writer.writerows(data)

    @staticmethod
    def get_distinct_values(data: list[dict], key: str) -> list:
        """
        Extracts distinct values for a given key from a list of dictionaries.

        Args:
            data (list): List of dictionaries.
            key (str): Key for which distinct values are needed.

        Returns:
            list: List of distinct values (excluding None).
        """
        return list({record.get(key) for record in data if key in record and record.get(key) is not None})

    @staticmethod
    def filter_list_of_dicts(
        data: List[Dict[str, Any]],
        attribute: str,
        value: Any,
        comparator: str = '==',
    ) -> List[Dict[str, Any]]:
        """
        Filters a list of dictionaries based on a comparator applied to a specific attribute.

        Args:
            data (list): List of dictionaries.
            attribute (str): Key to filter on.
            value (Any): Value to compare against.
            comparator (str): Comparison operator as a string (e.g., '==', '!=', '>', '<', '>=', '<=', 'in', 'not in').


        Returns:
            list: Filtered list of dictionaries.
        """

        # Supported operators
        ops: dict[str, Callable] = {
            '==': operator.eq,
            '!=': operator.ne,
            '>': operator.gt,
            '<': operator.lt,
            '>=': operator.ge,
            '<=': operator.le,
            'in': lambda a, b: a in b,
            'not in': lambda a, b: a not in b
        }

        if comparator not in ops:
            raise ValueError(f"Unsupported comparator '{comparator}'. Use one of: {list(ops.keys())}")

        func = ops[comparator]

        return [
            item for item in data
            if attribute in item and func(item[attribute], value)
        ]


import unittest
from pathlib import Path
from surquest.utils.appstoreconnect.credentials import Credentials
from surquest.utils.appstoreconnect.analyticsreports.client import Client
from surquest.utils.appstoreconnect.analyticsreports.enums.granularity import (
    Granularity,
)
from surquest.utils.appstoreconnect.analyticsreports.enums.report_name import ReportName


ISSUER_ID = "69a6de80-fd44-47e3-e053-5b8c7c11a4d1"
KEY_ID = "5WDUV3USAU"
PRIVATE_KEY_PATH = Path.cwd() / "credentials" / "key.p8"
PRIVATE_KEY = PRIVATE_KEY_PATH.read_text()
APP_ID = "950949627"
REPORT_NAME = ReportName.APP_STORE_INSTALLATION_AND_DELETION_STANDARD
GRANULARITY = Granularity.DAILY
DATE = "2025-07-27"


class TestClientIntegration(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.client = Client(
            credentials=Credentials(
                issuer_id=ISSUER_ID, key_id=KEY_ID, private_key=PRIVATE_KEY
            )
        )
        cls.test_app_id = APP_ID

    def test_read_report_requests(self):
        response = self.client.read_report_requests(self.test_app_id)
        assert isinstance(response, list), f"Expected list, got {type(response)}"
        if response:
            assert (
                "id" in response[0]
            ), "Each report request should contain an 'id' field"

    def test_read_report_for_specific_request(self):

        report_requests = self.client.read_report_requests(self.test_app_id)
        assert isinstance(
            report_requests, list
        ), f"Expected list, got {type(report_requests)}"
        assert report_requests, "No report requests found"

        request_id = report_requests[0]["id"]
        report_response = self.client.read_report_for_specific_request(request_id)
        assert isinstance(
            report_response, list
        ), f"Expected list, got {type(report_response)}"
        if report_response:
            assert (
                "id" in report_response[0]
            ), "Report item should contain an 'id' field"

    def test_read_list_of_instances_of_report(self):
        report_requests = self.client.read_report_requests(self.test_app_id)
        assert report_requests, "No report requests found"

        request_id = report_requests[0]["id"]
        report_response = self.client.read_report_for_specific_request(request_id)
        assert report_response, "No reports found for the given request"

        report_id = report_response[0]["id"]
        instances = self.client.read_list_of_instances_of_report(report_id)
        assert isinstance(instances, list), f"Expected list, got {type(instances)}"
        if instances:
            assert "id" in instances[0], "Instance item should contain an 'id' field"

    def test_download_report_to_dicts_with_invalid_url(self):
        data = self.client.download_report_to_dicts("https://invalid-url.com/report.gz")
        assert data is None, "Expected None for invalid URL"

    def test_list_report_dates(self):

        dates = self.client.list_report_dates(
            app_id=self.test_app_id,
            report_name=REPORT_NAME,
            granularity=GRANULARITY,
        )

        assert isinstance(dates, list), f"Expected list, got {type(dates)}"
        if dates:
            assert isinstance(
                dates[0], str
            ), f"Expected string in dates list, got {type(dates[0])}"


    def test_get_data_pipeline(self):
        data = self.client.get_data(
            app_id=self.test_app_id,
            report_name=REPORT_NAME,
            granularity=GRANULARITY,
            dates={DATE},
        )
        assert isinstance(data, list), f"Expected list, got {type(data)}"
        if data:
            assert isinstance(
                data[0], dict
            ), f"Expected dict in data rows, got {type(data[0])}"

    def test__get_request(self):

        response = self.client._get_request("https://httpstat.us/400")
        assert None == response
        
    def test_fetch_customer_reviews(self):
        reviews = self.client.fetch_customer_reviews(
            app_id=self.test_app_id,
            max_iterations=2,  # Limit for test efficiency
        )

        assert isinstance(reviews, list), f"Expected list, got {type(reviews)}"
        if reviews:
            first = reviews[0]
            assert isinstance(first, dict), f"Expected dict, got {type(first)}"
            assert "id" in first, "Each review should contain an 'id' field"


from enum import Enum

class Category(Enum):
    APP_USAGE = "APP_USAGE"
    APP_STORE_ENGAGEMENT = "APP_STORE_ENGAGEMENT"
    COMMERCE = "COMMERCE"
    FRAMEWORK_USAGE = "FRAMEWORK_USAGE"
    PERFORMANCE = "PERFORMANCE"